@@ -17,18 +17,16 @@
  */
 
 #include "config.h"
+#include "ass_compat.h"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <strings.h>
 #include <assert.h>
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <unistd.h>
 #include <inttypes.h>
-#include <ctype.h>
 
 #ifdef CONFIG_ICONV
 #include <iconv.h>
@@ -37,6 +35,7 @@
 #include "ass.h"
 #include "ass_utils.h"
 #include "ass_library.h"
+#include "ass_string.h"
 
 #define ass_atof(STR) (ass_strtod((STR),NULL))
 
@@ -54,6 +53,11 @@ struct parser_priv {
     char *fontdata;
     int fontdata_size;
     int fontdata_used;
+
+    // contains bitmap of ReadOrder IDs of all read events
+    uint32_t *read_order_bitmap;
+    int read_order_elems; // size in uint32_t units of read_order_bitmap
+    int check_readorder;
 };
 
 #define ASS_STYLES_ALLOC 20
@@ -68,6 +72,7 @@ void ass_free_track(ASS_Track *track)
     int i;
 
     if (track->parser_priv) {
+        free(track->parser_priv->read_order_bitmap);
         free(track->parser_priv->fontname);
         free(track->parser_priv->fontdata);
         free(track->parser_priv);
@@ -151,6 +156,45 @@ void ass_free_style(ASS_Track *track, int sid)
     free(style->FontName);
 }
 
+static int resize_read_order_bitmap(ASS_Track *track, int max_id)
+{
+    // Don't allow malicious files to OOM us easily. Also avoids int overflows.
+    if (max_id < 0 || max_id >= 10 * 1024 * 1024 * 8)
+        goto fail;
+    if (max_id >= track->parser_priv->read_order_elems * 32) {
+        int oldelems = track->parser_priv->read_order_elems;
+        int elems = ((max_id + 31) / 32 + 1) * 2;
+        assert(elems >= oldelems);
+        track->parser_priv->read_order_elems = elems;
+        void *new_bitmap =
+            realloc(track->parser_priv->read_order_bitmap, elems * 4);
+        if (!new_bitmap)
+            goto fail;
+        track->parser_priv->read_order_bitmap = new_bitmap;
+        memset(track->parser_priv->read_order_bitmap + oldelems, 0,
+               (elems - oldelems) * 4);
+    }
+    return 0;
+
+fail:
+    free(track->parser_priv->read_order_bitmap);
+    track->parser_priv->read_order_bitmap = NULL;
+    track->parser_priv->read_order_elems = 0;
+    return -1;
+}
+
+static int test_and_set_read_order_bit(ASS_Track *track, int id)
+{
+    if (resize_read_order_bitmap(track, id) < 0)
+        return -1;
+    int index = id / 32;
+    int bit = 1 << (id % 32);
+    if (track->parser_priv->read_order_bitmap[index] & bit)
+        return 1;
+    track->parser_priv->read_order_bitmap[index] |= bit;
+    return 0;
+}
+
 // ==============================================================================================
 
 /**
@@ -181,14 +225,14 @@ static void set_default_style(ASS_Style *style)
 
 static long long string2timecode(ASS_Library *library, char *p)
 {
-    unsigned h, m, s, ms;
+    int h, m, s, ms;
     long long tm;
-    int res = sscanf(p, "%1d:%2d:%2d.%2d", &h, &m, &s, &ms);
+    int res = sscanf(p, "%d:%d:%d.%d", &h, &m, &s, &ms);
     if (res < 4) {
         ass_msg(library, MSGL_WARN, "Bad timestamp");
         return 0;
     }
-    tm = ((h * 60 + m) * 60 + s) * 1000 + ms * 10;
+    tm = ((h * 60LL + m) * 60 + s) * 1000 + ms * 10;
     return tm;
 }
 
@@ -212,7 +256,7 @@ static int numpad2align(int val)
 
 
 #define ALIAS(alias,name) \
-        if (strcasecmp(tname, #alias) == 0) {tname = #name;}
+        if (ass_strcasecmp(tname, #alias) == 0) {tname = #name;}
 
 /* One section started with PARSE_START and PARSE_END parses a single token
  * (contained in the variable named token) for the header indicated by the
@@ -228,32 +272,29 @@ static int numpad2align(int val)
 #define PARSE_END   }
 
 #define ANYVAL(name,func) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		target->name = func(token);
 
 #define STRVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		if (target->name != NULL) free(target->name); \
 		target->name = strdup(token);
 
 #define STARREDSTRVAL(name) \
-    } else if (strcasecmp(tname, #name) == 0) { \
+    } else if (ass_strcasecmp(tname, #name) == 0) { \
         if (target->name != NULL) free(target->name); \
         while (*token == '*') ++token; \
         target->name = strdup(token);
 
-#define COLORVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
-		target->name = string2color(track->library, token, 0);
-
+#define COLORVAL(name) ANYVAL(name,parse_color_header)
 #define INTVAL(name) ANYVAL(name,atoi)
 #define FPVAL(name) ANYVAL(name,ass_atof)
 #define TIMEVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		target->name = string2timecode(track->library, token);
 
 #define STYLEVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		target->name = lookup_style(track, token);
 
 static char *next_token(char **str)
@@ -312,7 +353,7 @@ static int process_event_tail(ASS_Track *track, ASS_Event *event,
 
     while (1) {
         NEXT(q, tname);
-        if (strcasecmp(tname, "Text") == 0) {
+        if (ass_strcasecmp(tname, "Text") == 0) {
             char *last;
             event->Text = strdup(p);
             if (*event->Text != 0) {
@@ -365,19 +406,19 @@ void ass_process_force_style(ASS_Track *track)
         *eq = '\0';
         token = eq + 1;
 
-        if (!strcasecmp(*fs, "PlayResX"))
+        if (!ass_strcasecmp(*fs, "PlayResX"))
             track->PlayResX = atoi(token);
-        else if (!strcasecmp(*fs, "PlayResY"))
+        else if (!ass_strcasecmp(*fs, "PlayResY"))
             track->PlayResY = atoi(token);
-        else if (!strcasecmp(*fs, "Timer"))
+        else if (!ass_strcasecmp(*fs, "Timer"))
             track->Timer = ass_atof(token);
-        else if (!strcasecmp(*fs, "WrapStyle"))
+        else if (!ass_strcasecmp(*fs, "WrapStyle"))
             track->WrapStyle = atoi(token);
-        else if (!strcasecmp(*fs, "ScaledBorderAndShadow"))
+        else if (!ass_strcasecmp(*fs, "ScaledBorderAndShadow"))
             track->ScaledBorderAndShadow = parse_bool(token);
-        else if (!strcasecmp(*fs, "Kerning"))
+        else if (!ass_strcasecmp(*fs, "Kerning"))
             track->Kerning = parse_bool(token);
-        else if (!strcasecmp(*fs, "YCbCr Matrix"))
+        else if (!ass_strcasecmp(*fs, "YCbCr Matrix"))
             track->YCbCrMatrix = parse_ycbcr_matrix(token);
 
         dt = strrchr(*fs, '.');
@@ -391,7 +432,7 @@ void ass_process_force_style(ASS_Track *track)
         }
         for (sid = 0; sid < track->n_styles; ++sid) {
             if (style == NULL
-                || strcasecmp(track->styles[sid].Name, style) == 0) {
+                || ass_strcasecmp(track->styles[sid].Name, style) == 0) {
                 target = track->styles + sid;
                 PARSE_START
                     STRVAL(FontName)
@@ -578,7 +619,7 @@ static int process_info_line(ASS_Track *track, char *str)
         track->YCbCrMatrix = parse_ycbcr_matrix(str + 13);
     } else if (!strncmp(str, "Language:", 9)) {
         char *p = str + 9;
-        while (*p && isspace(*p)) p++;
+        while (*p && ass_isspace(*p)) p++;
         track->Language = strndup(p, 2);
     }
     return 0;
@@ -746,17 +787,17 @@ static int process_fonts_line(ASS_Track *track, char *str)
 */
 static int process_line(ASS_Track *track, char *str)
 {
-    if (!strncasecmp(str, "[Script Info]", 13)) {
+    if (!ass_strncasecmp(str, "[Script Info]", 13)) {
         track->parser_priv->state = PST_INFO;
-    } else if (!strncasecmp(str, "[V4 Styles]", 11)) {
+    } else if (!ass_strncasecmp(str, "[V4 Styles]", 11)) {
         track->parser_priv->state = PST_STYLES;
         track->track_type = TRACK_TYPE_SSA;
-    } else if (!strncasecmp(str, "[V4+ Styles]", 12)) {
+    } else if (!ass_strncasecmp(str, "[V4+ Styles]", 12)) {
         track->parser_priv->state = PST_STYLES;
         track->track_type = TRACK_TYPE_ASS;
-    } else if (!strncasecmp(str, "[Events]", 8)) {
+    } else if (!ass_strncasecmp(str, "[Events]", 8)) {
         track->parser_priv->state = PST_EVENTS;
-    } else if (!strncasecmp(str, "[Fonts]", 7)) {
+    } else if (!ass_strncasecmp(str, "[Fonts]", 7)) {
         track->parser_priv->state = PST_FONTS;
     } else {
         switch (track->parser_priv->state) {
@@ -850,13 +891,20 @@ void ass_process_codec_private(ASS_Track *track, char *data, int size)
 
 static int check_duplicate_event(ASS_Track *track, int ReadOrder)
 {
-    int i;
-    for (i = 0; i < track->n_events - 1; ++i)   // ignoring last event, it is the one we are comparing with
+    if (track->parser_priv->read_order_bitmap)
+        return test_and_set_read_order_bit(track, ReadOrder) > 0;
+    // ignoring last event, it is the one we are comparing with
+    for (int i = 0; i < track->n_events - 1; i++)
         if (track->events[i].ReadOrder == ReadOrder)
             return 1;
     return 0;
 }
 
+void ass_set_check_readorder(ASS_Track *track, int check_readorder)
+{
+    track->parser_priv->check_readorder = check_readorder == 1;
+}
+
 /**
  * \brief Process a chunk of subtitle stream data. In Matroska, this contains exactly 1 event (or a commentary).
  * \param track track
@@ -873,6 +921,14 @@ void ass_process_chunk(ASS_Track *track, char *data, int size,
     char *p;
     char *token;
     ASS_Event *event;
+    int check_readorder = track->parser_priv->check_readorder;
+
+    if (check_readorder && !track->parser_priv->read_order_bitmap) {
+        for (int i = 0; i < track->n_events; i++) {
+            if (test_and_set_read_order_bit(track, track->events[i].ReadOrder) < 0)
+                break;
+        }
+    }
 
     if (!track->event_format) {
         ass_msg(track->library, MSGL_WARN, "Event format header missing");
@@ -895,7 +951,7 @@ void ass_process_chunk(ASS_Track *track, char *data, int size,
     do {
         NEXT(p, token);
         event->ReadOrder = atoi(token);
-        if (check_duplicate_event(track, event->ReadOrder))
+        if (check_readorder && check_duplicate_event(track, event->ReadOrder))
             break;
 
         NEXT(p, token);
@@ -928,6 +984,9 @@ void ass_flush_events(ASS_Track *track)
             ass_free_event(track, eid);
         track->n_events = 0;
     }
+    free(track->parser_priv->read_order_bitmap);
+    track->parser_priv->read_order_bitmap = NULL;
+    track->parser_priv->read_order_elems = 0;
 }
 
 #ifdef CONFIG_ICONV
@@ -945,31 +1004,12 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
     char *outbuf;
     assert(codepage);
 
-    {
-        const char *cp_tmp = codepage;
-#ifdef CONFIG_ENCA
-        char enca_lang[3], enca_fallback[100];
-        if (sscanf(codepage, "enca:%2s:%99s", enca_lang, enca_fallback) == 2
-            || sscanf(codepage, "ENCA:%2s:%99s", enca_lang,
-                      enca_fallback) == 2) {
-            cp_tmp =
-                ass_guess_buffer_cp(library, (unsigned char *) data, size,
-                                    enca_lang, enca_fallback);
-        }
-#endif
-        if ((icdsc = iconv_open(tocp, cp_tmp)) != (iconv_t) (-1)) {
-            ass_msg(library, MSGL_V, "Opened iconv descriptor");
-        } else
-            ass_msg(library, MSGL_ERR, "Error opening iconv descriptor");
-#ifdef CONFIG_ENCA
-        if (cp_tmp != codepage) {
-            free((void*)cp_tmp);
-        }
-#endif
-    }
-
-    if (icdsc == (iconv_t) (-1))
+    if ((icdsc = iconv_open(tocp, codepage)) != (iconv_t) (-1)) {
+        ass_msg(library, MSGL_V, "Opened iconv descriptor");
+    } else {
+        ass_msg(library, MSGL_ERR, "Error opening iconv descriptor");
         return NULL;
+    }
 
     {
         size_t osize = size;
@@ -1021,7 +1061,6 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
 out:
     if (icdsc != (iconv_t) (-1)) {
         (void) iconv_close(icdsc);
-        icdsc = (iconv_t) (-1);
         ass_msg(library, MSGL_V, "Closed iconv descriptor");
     }
 
@@ -1035,7 +1074,7 @@ out:
  * \param bufsize out: file size
  * \return pointer to file contents. Caller is responsible for its deallocation.
  */
-static char *read_file(ASS_Library *library, char *fname, size_t *bufsize)
+char *read_file(ASS_Library *library, char *fname, size_t *bufsize)
 {
     int res;
     long sz;
@@ -1232,12 +1271,13 @@ int ass_read_styles(ASS_Track *track, char *fname, char *codepage)
         buf = tmpbuf;
     }
     if (!buf)
-        return 0;
+        return 1;
 #endif
 
     old_state = track->parser_priv->state;
     track->parser_priv->state = PST_STYLES;
     process_text(track, buf);
+    free(buf);
     track->parser_priv->state = old_state;
 
     return 0;
@@ -1248,14 +1288,12 @@ long long ass_step_sub(ASS_Track *track, long long now, int movement)
     int i;
     ASS_Event *best = NULL;
     long long target = now;
-    int direction = movement > 0 ? 1 : -1;
+    int direction = (movement > 0 ? 1 : -1) * !!movement;
 
-    if (movement == 0)
-        return 0;
     if (track->n_events == 0)
         return 0;
 
-    while (movement) {
+    do {
         ASS_Event *closest = NULL;
         long long closest_time = now;
         for (i = 0; i < track->n_events; i++) {
@@ -1268,7 +1306,7 @@ long long ass_step_sub(ASS_Track *track, long long now, int movement)
                         closest_time = end;
                     }
                 }
-            } else {
+            } else if (direction > 0) {
                 long long start = track->events[i].Start;
                 if (start > target) {
                     if (!closest || start < closest_time) {
@@ -1276,13 +1314,21 @@ long long ass_step_sub(ASS_Track *track, long long now, int movement)
                         closest_time = start;
                     }
                 }
+            } else {
+                long long start = track->events[i].Start;
+                if (start < target) {
+                    if (!closest || start >= closest_time) {
+                        closest = &track->events[i];
+                        closest_time = start;
+                    }
+                }
             }
         }
         target = closest_time + direction;
         movement -= direction;
         if (closest)
             best = closest;
-    }
+    } while (movement);
 
     return best ? best->Start - now : 0;
 }
@@ -1299,6 +1345,7 @@ ASS_Track *ass_new_track(ASS_Library *library)
         free(track);
         return NULL;
     }
+    track->parser_priv->check_readorder = 1;
     return track;
 }
 
